#!/usr/bin/env python
# This file is automatically generated by pywf script
# based on the content in pyproject.toml file

__version__ = "0.1.3"
__short_description__ = "A type-safe hierarchical configuration management system for AWS projects that provides environment-specific inheritance, validation, and integrated deployment to AWS SSM Parameter Store and S3 with custom versioning."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@email.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@email.com"

# run this script to print out the version number
if __name__ == "__main__":  # pragma: no cover
    print(__version__)
